#include <windows.h>

#include <dinput.h>

#include <stdio.h>
#include <stdlib.h>

struct key_entry_t {
	const char *	name;
	unsigned int	address;
	unsigned int	orig_address;
	unsigned int	key;
};

struct key_entry_t key_table[] = {
	{ "KeyUp",	0x56a9,		0x56a9,		DIK_UP		},
	{ "KeyDown",	0x56ba,		0x56ba,		DIK_DOWN	},
	{ "KeyLeft",	0x56cb,		0x56cb,		DIK_LEFT	},
	{ "KeyRight",	0x56dc,		0x56dc,		DIK_RIGHT	},
	{ "KeyA",	0x56ed,		0x56ed,		DIK_Z		},
	{ "KeyB",	0x56fe,		0x56fe,		DIK_X		},
	{ "KeyC",	0x570f,		0x570f,		DIK_C		},
	{ "KeyTouch",	0x5720,		0x5720,		DIK_S		},
	{ "KeyAppeal",	0x5731,		0x5731,		DIK_LSHIFT	},
	{ "KeyUnknown1",0x573f,		0x573f,		DIK_A		},
	{ "KeyUnknown2",0x5750,		0x5750,		DIK_D		},
	{ "KeyUnknown3",0x5761,		0x5761,		DIK_Q		},
	{ "KeyUnknown4",0x5772,		0x5772,		DIK_E		},
	{ "KeyStart",	0x5783,		0x5783,		DIK_W		},
	{ "KeyUnknown5",0x5794,		0x5794,		DIK_RETURN	},
	{ "KeyUnknown6",0x5a495,	0x57a5,		DIK_RSHIFT	},
	{ "KeyAB",	0x5a462,	0x0,		0		},
	{ "KeyBC",	0x5a473,	0x0,		0		},
	{ "KeyABC",	0x5a484,	0x0,		0		},
};

const unsigned char code_1[7] = {
	// JMP 45b060
	0xe9, 0xb8, 0x4c, 0x05, 0x00, 0x90, 0x90
};
const unsigned int code_1_size = sizeof(code_1);
const unsigned int code_1_address = 0x57a3;

const unsigned char code_2[] = {
	// CMP BYTE PTR SS:[EBP+xx], 0
	0x80, 0xbd, 0x00, 0x00, 0x00, 0x00, 0x00,
	// JE SHORT blahblah
	0x74, 0x08,
	// OR DWORD PTR SS:[ESP+14], 20 (AB)
	0x81, 0x4c, 0x24, 0x14, 0x20, 0x00, 0x00, 0x00,
	
	// again, BC (0x40)
	0x80, 0xbd, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x74, 0x08,
	0x81, 0x4c, 0x24, 0x14, 0x40, 0x00, 0x00, 0x00,
	
	// again, ABC (0x80)
	0x80, 0xbd, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x74, 0x08,
	0x81, 0x4c, 0x24, 0x14, 0x80, 0x00, 0x00, 0x00,
	
	// CMP BYTE PTR SS:[EBP+75A], 0
	0x80, 0xbd, 0x5a, 0x07, 0x00, 0x00, 0x00,
	// JMP 4063AA 					FIXME
	0xe9, 0x0b, 0xb3, 0xfa, 0xff
};
const unsigned int code_2_size = sizeof(code_2);

const unsigned int code_2_address = 0x5a460;	// 45b060

int key_table_size = sizeof(key_table)/sizeof(key_table[0]);

void run() {
	FILE *file = fopen("GameMain.exe", "rb");
	unsigned char *data;
	unsigned int size;
	int i;
	
	if (!file) {
		printf("Could not open GameMain.exe\n");
		
		return;
	}
	
	fseek(file, 0, SEEK_END);
	size = ftell(file);
	fseek(file, 0, SEEK_SET);
	
	if (size < 0xb0000) {
		printf("Not a valid executable.\n");
		
		return;
	}
	
	data = malloc(size);
	fread(data, size, 1, file);
	
	fclose(file);
	
	/* verify that it is valid. */
	if (memcmp(data + 0x5dbce, " v1.01", 6) != 0) {
		printf("This version of kbrebind only works with v1.01.\n");
		
		free(data);
		return;
	}
	
	/* get all original keys */
	for (i = 0; i < key_table_size; ++i) {
		if (key_table[i].orig_address == 0) {
			continue;
		}
		
		key_table[i].key = *(unsigned int *)(data + key_table[i].orig_address) - 0x724;
	}
	
	/* write new keybind code */
	memcpy(data + code_1_address, code_1, code_1_size);
	memcpy(data + code_2_address, code_2, code_2_size);
	
	/* update binds */
	for (i = 0; i < key_table_size; ++i) {
		unsigned int orig;
		if (key_table[i].address == 0) {
			continue;
		}
		
		orig = key_table[i].key;
		
		key_table[i].key = GetPrivateProfileInt("KEYS", key_table[i].name, key_table[i].key, "./kbrebind.ini");
		
		*(unsigned int *)(data + key_table[i].address) = key_table[i].key + 0x724;
		printf("%20s: %8x -> %8x\n", key_table[i].name, orig, *(unsigned int *)(data + key_table[i].address) - (0x733 - 0x1f));
	}
	
	file = fopen("GameMain-kb.exe", "wb");
	if (file) {
		fwrite(data, size, 1, file);
		fclose(file);
	}
	
	printf("done! GameMain-kb.exe created\n");
}

int main() {
	run();
	
	printf("press enter to close.\n");
	getchar();
	
	return 0;
}
